from datamana.ipc import Semaphore, MQueue, SharedMemory
import numpy as np

class Base():
    def __init__(self, name):
        self.data_name = f'datamana{name}'
        self.data_meta_name = f'{self.data_name}meta'
        self.data_share_name = f'{self.data_name}share'
        self.data_sem_name = f'{self.data_name}sem'
        self.data_event_name =f'{self.data_name}event'
        self.sem = Semaphore(self.data_sem_name)
        self.event = Semaphore(self.data_event_name)
        self.name2shm = {}

    def get_shm(self, shm_name, size, **kwargs):
        if shm_name in self.name2shm:
            shm = self.name2shm[shm_name]
            if size > 0: shm.resize(size)
        else:
            shm = SharedMemory(name=shm_name, size=size, **kwargs)
            self.name2shm[shm_name] = shm
        return shm

    def write_byte(self, shm, data, size):
        shm.resize(size)
        shm.buf[:size] = data[:]

    def write_numpy(self, shm, data):
        np_shm = np.ndarray(data.shape, dtype=data.dtype, buffer=shm.buf)
        np_shm[:] = data[:]



import os, mmap
from datamana._C import (
    Semaphore as __Semaphore,
    MQueue as __MQueue,
    SharedMemory as __SharedMemory,
)

class Semaphore(__Semaphore):
    def __init__(self, name, oflag=os.O_CREAT | os.O_RDWR, mode=0o600, value=1):
        super().__init__()
        errno = self.open(name, oflag, mode, value)
        if errno != 0:
            raise RuntimeError(f'create Semaphore failed! <{os.strerror(errno)}>')

del __Semaphore

class MQueue(__MQueue):
    def __init__(self, name, oflag=os.O_CREAT | os.O_RDWR, mode=0o600, msgsize=1, maxmsg=1):
        super().__init__()
        self.msgsize = msgsize
        self.maxmsg = maxmsg
        name = '/' + name
        errno = self.open(name, oflag, mode)
        if errno != 0:
            raise RuntimeError(f'create MQueue failed! <{os.strerror(errno)}>')

del __MQueue

class SharedMemory(__SharedMemory):
    def __init__(self, name, size, oflag=os.O_CREAT | os.O_RDWR, mode=0o600):
        if size < 0:
            raise ValueError("'size' must be a positive integer")
        super().__init__()
        name = '/' + name
        errno = self.open(name, oflag, mode)
        if errno != 0:
            raise RuntimeError(f'create SharedMemory failed! <{os.strerror(errno)}>')
        if oflag & os.O_CREAT:
            self.resize(size)
        self._map_size = 0

    def resize(self, size, strict=False):
        shm_size = self.shm_size
        if shm_size < size or (shm_size > size and strict):
            os.ftruncate(self.fd, size)

    def __remap(self, size):
        self._mmap = mmap.mmap(self.fd, size)
        self._buf = memoryview(self._mmap)
        self._map_size = size

    @property
    def shm_size(self):
        stats = os.fstat(self.fd)
        return stats.st_size

    @property
    def buf(self):
        shm_size = self.shm_size
        if shm_size != self._map_size:
            self.__remap(shm_size)
        return self._buf

del __SharedMemory



import os, time, pickle
from typing import Iterable
from datamana.base import Base
import numpy as np

class Server(Base):
    def __init__(self, name, dataloader: Iterable):
        super().__init__(name)
        self.dataloader = dataloader
        self.iterloader = iter(self.dataloader)
        self.oflags = os.O_CREAT | os.O_RDWR
        self.data_id = 0
        self.max_data_id = 2**31 - 1

    def next(self):
        try:
            data = next(self.iterloader)
        except StopIteration:
            self.iterloader = iter(self.dataloader)
            data = next(self.iterloader)
        shm = self.get_shm(self.data_share_name, data.nbytes, oflag=self.oflags)
        self.write_numpy(shm, data)

        meta_data = {
            'shape': data.shape,
            'dtype': data.dtype.name,
            'id': self.data_id,
        }
        meta_data_pkl = pickle.dumps(meta_data)
        pkl_size = len(meta_data_pkl)
        shm = self.get_shm(self.data_meta_name, pkl_size, oflag=self.oflags)
        self.write_byte(shm, meta_data_pkl, pkl_size)
        self.data_id += 1
        if self.data_id >= self.max_data_id:
            self.data_id = 0

    def serve(self):
        self.next()

        while True:
            # wait client signal
            ret = self.event.wait()
            loop = ret
            # clean all signal
            while loop == 0:
                loop = self.event.trywait()

            if ret == 0:
                self.sem.wait()
                self.next()
                self.sem.post()
            else:
                print(os.strerror(ret))
                time.sleep(1)

class Client(Base):
    def __init__(self, name):
        super().__init__(name)
        self.pid = os.getpid()
        self.oflags = os.O_RDWR
        self.data_id = -1

    def next(self):
        while True:
            self.sem.wait()

            shm_meta = self.get_shm(self.data_meta_name, 0, oflag=self.oflags)
            meta_data = pickle.loads(shm_meta.buf)

            if self.data_id != meta_data['id']:
                self.data_id = meta_data['id']
                shape = meta_data['shape']
                dtype = np.dtype(meta_data['dtype'])

                shm_data = self.get_shm(self.data_share_name, 0, oflag=self.oflags)
                data = np.ndarray(shape, dtype=dtype, buffer=shm_data.buf).copy()

                self.sem.post()
                return data
            else:
                self.sem.post()
                self.event.post()
